;;; "kruskal-gallenbachergraph.scm"
(load "breakpoint.scm")
; Programm zum Algorithmus von Kruskal
; Loest das Problem des minimal-spanning-tree mit dem
; Algorithmus von kruskal

; Beginne mit einer Liste der Teilbaeume, die jeweils
; genau einen Knoten enthalten
; Fuege dann alle Kanten nach den Bewertungen geordnet hinzu,
; die genau zwei Teilbaeume verbinden,
; bis die Liste der Teilbaeume genau ein Element enthaelt

;;; ===== Datenstrukturen =====
; Kanten: (<knoten-1> <knoten-2> <bewertung>)
; teilbaum: (<liste der knoten> <liste der kanten>)

;;;====== Praedikate ==========================================
;;; ----- vor? -----
(define
  (vor? kante-1 kante-2)
  (< (third kante-1) (third kante-2)))

; ----- knoten-im-teilbaum? -----
; prueft, ob ein knoten in einem teilbaum enthalten ist.
(define 
  (knoten-im-teilbaum? knoten teilbaum)
  (member knoten (first teilbaum)))

(define
  (beide-knoten-im-teilbaum? kante teilbaum)
  (and
   (knoten-im-teilbaum? (first kante) teilbaum)
   (knoten-im-teilbaum? (second kante) teilbaum)))

; ----- zyklus? -----
; eine neue Kante bildet einen Zyklus, wenn beide Knoten im selben
; teilbaum enthalten sind
(define
  (zyklus? kante teilbaeume)
  (cond
    ((null? teilbaeume)
     #f)
    ((beide-knoten-im-teilbaum? kante (first teilbaeume))
     #t)
    (else
     (zyklus? kante (rest teilbaeume)))))
;;Tests:
;(zyklus? '(E D 3) '(((D E) ((D E 3))) ((F) ()) ((G) ()) ((H) ()) ((C) ()) ((B) ()) ((A) ())))
;(zyklus? '(A B 8) '(((D E) ((D E 3))) ((F) ()) ((G) ()) ((H) ()) ((C) ()) ((B) ()) ((A) ())))

;;; ===== Sortieren der Kanten nach den Kantenbewertungen =====
; wird mit Hilfe der Funktionen fuer die Prioritaetswarteschlange
; umgesetzt, die dafuer einmal am Anfang aufgerufenwerden.
(load "prioritaetswarteschlange.scm")
; Test:
;(fuege-alle-ein kanten vor? '())

;;; ===== erzeuge-teilbaeume =====
; erzeugt den Anfangszustand aus der Kantenliste
(define
  (erzeuge-teilbaeume kanten)
  (define
    (erzeuge-intern knoten teilbaeume kanten)
    (cond
      ((null? kanten)
       teilbaeume)
      ((member (first (first kanten)) knoten) ; hab ich schon
       (erzeuge-intern knoten teilbaeume (rest kanten)))
      (else ; sonst aufnehmen
       (erzeuge-intern
        (cons (first (first kanten)) knoten)
        (cons (list (list (first (first kanten))) '()) teilbaeume)
        (rest kanten)))
      ))
  (erzeuge-intern '() '() kanten)
  )
;(erzeuge-teilbaeume kanten)

;;; ===== weitere Hilfsfunktionen zur Uebersichtlichkeit =====

; ----- beide-nach-vorn -----
; bringt die beiden teilbaeume zur Kante nach vorn in die Liste
; der teilbaeume
(define
  (beide-nach-vorn kante teilbaeume)
  (define
    (beide-intern beide bearbeitet teilbaeume)
    ;(ausgabe beide bearbeitet teilbaeume)
    (cond
      ((= 2 (length beide))
       (append beide bearbeitet teilbaeume))
      ((knoten-im-teilbaum? (first kante) (first teilbaeume))
       (beide-intern
        (cons (first teilbaeume) beide)
        bearbeitet
        (rest teilbaeume)))
      ((knoten-im-teilbaum? (second kante) (first teilbaeume))
       (beide-intern
        (cons (first teilbaeume) beide)
        bearbeitet
        (rest teilbaeume)))
      (else
        (beide-intern
         beide
         (cons (first teilbaeume) bearbeitet)
         (rest teilbaeume)))
      ))
  (beide-intern '() '() teilbaeume))
; Test:
;(beide-nach-vorn '(D E 3) (erzeuge-teilbaeume kanten))

; ----- verbinde-zwei -----
; verbindet die beiden Teilbaeume mit der Kante zu einem
(define
  (verbinde-zwei kante ersten zweiten)
  (list
   ; knoten
   (append (first ersten) (first zweiten))
   ;kanten
   (cons kante (append (second ersten) (second zweiten)))))

; ----- verbinde-teilbaeume-mit-kante -----
; verbindet die beiden teilbaeume mit der kante
(define
  (verbinde-teilbaeume-mit-kante kante teilbaeume)
  (define
    (verbinde-intern kante teilbaeume)
    (cons
     (verbinde-zwei kante (first teilbaeume) (second teilbaeume))
     (rest (rest teilbaeume))))
  (verbinde-intern kante (beide-nach-vorn kante teilbaeume)))

;;; ===== kruskal =====
(define
  (kruskal kanten)
  (define
    (kruskal-intern kanten teilbaeume)
    ;(ausgabe  kanten teilbaeume)
    (cond
      ((null? (rest teilbaeume))
       (first teilbaeume))
      ((null? kanten)
       "Der Graph ist nicht zusammenhaengend!")
      ((zyklus? (first kanten) teilbaeume)
       (kruskal-intern (rest kanten) teilbaeume))
      (else
       (kruskal-intern
        (rest kanten)
        (verbinde-teilbaeume-mit-kante (first kanten) teilbaeume)))
      ))
  (kruskal-intern
   (fuege-alle-ein kanten vor? '())
   (erzeuge-teilbaeume kanten))
  )

; Tests:
(load "gallenbacher-graph-kanten.scm")
(kruskal *kanten*)
;;; Boeser Graph:
;(define
;  *boeser-Graph*
;  '((A B 2) (B A 2) (B C 10) (C B 10) (C D 1) (D C 1) (X Y 1) (Y X 1)))
;(kruskal *boeser-Graph*)